#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <algorithm>
#include "getq.h"
#include "basicdefs.h"

using namespace std;

// one-locus ordered Ewens Sampling Formula / PIM counterpart
double ESF(vector<int> & vec, double theta, vector<double>& M)
{
	double tmp = 1.0;
	int add = 0;
	int factor;
	
	for (int i=0; i < vec.size(); i++)
	{
		if (vec[i] < 0)
			return 0.0;
	}
	
	if (M.empty())
	{
		// Product[(n[[i]]-1)!,{i,1,Length[n]}]*t^Length[n]/Product[t+i,{i,0,Total[n]-1}]
		// t^k/(t(t+1)(t+2)...(t+k-1))
		for(int i=0; i < vec.size(); i++){
			if(vec[i] > 0){
				tmp *= theta/(theta+add);
				add += 1;
			}
		}
		// prod_i (n_i - 1)! and the rest of the denominator
		for(int i=0; i < vec.size(); i++){
			if(vec[i] > 1){
				factor = vec[i]-1;
				for(int j=1; j < vec[i]; j++){
					tmp *= factor/(theta+add);
					factor -= 1;
					add += 1;
				}
			}
		}
	}
	else
	{
		for(int i=0; i < vec.size(); i++){
			if(vec[i] > 0){
				factor = 0;
				for(int j=0; j < vec[i]; j++){
					tmp *= (factor + theta*M[i])/(theta+add);
					factor++;
					add++;
				}
			}
		}
	}
	return tmp;
}

double qzero(vector<int>& aVec, vector<int>& bVec, vector<vector<int> >& cMat,double thetaA, double thetaB, vector<double>& P, vector<double>& Q){
  vector<int> cA = marginalVec(cMat,'A');
  vector<int> cB = marginalVec(cMat,'B');
  vector<int> a_cA = aVec+cA;
  vector<int> b_cB = bVec+cB;
  return ESF(a_cA,thetaA,P)*ESF(b_cB,thetaB,Q);
}

double qone(vector<int>& aVec, vector<int>& bVec, vector<vector<int> >& cMat,double thetaA, double thetaB, vector<double>& P, vector<double>& Q){
  vector<int> cA = marginalVec(cMat,'A');
  vector<int> cB = marginalVec(cMat,'B');
  vector<int> a_cA = aVec+cA;
  vector<int> b_cB = bVec+cB;
  int c = total(cA);
  double tmp=0.0, result=c*(c-1)*ESF(a_cA,thetaA,P)*ESF(b_cB,thetaB,Q);
  int i, j;
  for(i=0; i < aVec.size(); i++){
    a_cA[i] -= 1;
    tmp += ESF(a_cA,thetaA,P)*cA[i]*(cA[i]-1);
    a_cA[i] += 1;
  }
  result += -ESF(b_cB,thetaB,Q)*tmp;

  tmp = 0.0;
  for(j=0; j < bVec.size(); j++){
    b_cB[j] -= 1;
    tmp += ESF(b_cB,thetaB,Q)*cB[j]*(cB[j]-1);
    b_cB[j] += 1;
  }
  result += -ESF(a_cA,thetaA,P)*tmp;
  tmp = 0.0;
  for(i=0; i < cMat.size(); i++){
    a_cA[i] -= 1;
    for(j=0; j < cMat[0].size(); j++){
      b_cB[j] -= 1;
      tmp += cMat[i][j]*(cMat[i][j]-1)*ESF(a_cA,thetaA,P)*ESF(b_cB,thetaB,Q);
      b_cB[j] += 1;
    }
    a_cA[i] += 1;
  }

  result += tmp;
  return 0.5*result;
}
  


double qtwoEg(vector<int>& aVec, vector<int>& bVec, vector<vector<int> >& cMat,double tA, double tB, vector<double>& P, vector<double>& Q){
  vector<int> cA = marginalVec(cMat,'A');
  vector<int> cB = marginalVec(cMat,'B');
  double K = cMat.size();
  double L = cMat[0].size();
  double a = total(aVec);
  double b = total(bVec);
  double c = total(cA);
  
	vector<double> term(29);  // term[0] is not used.
	vector<int> a_cA = aVec+cA;
	vector<int> b_cB = bVec+cB;
	vector<int> tempmatrix,tempmatrix2;
	int i,j,k,l;
	fill(term.begin(),term.end(),0.0);
	
	term[1] = c*(c+1)*(3*c-2)*(c-1)*ESF(a_cA,tA,P)*ESF(b_cB,tB,Q)/24.0;
	
	for(i=0; i< K; i++){
		tempmatrix = aVec+cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		term[2] = term[2] - cA[i]*(cA[i]-1)*(c+1)*(3*c-2)*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q)/24.0;
	}
	
	for(j=0; j< L; j++){
		tempmatrix = bVec+cB;
		tempmatrix[j] = tempmatrix[j] - 1;
		term[3] = term[3] - cB[j]*(cB[j]-1)*(c+1)*(3*c-2)*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q)/24.0;
	}
	
	
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		for(j=0; j< L; j++){
			tempmatrix2 = b_cB;
			tempmatrix2[j] = tempmatrix2[j] - 1;
			term[4] = term[4] + cMat[i][j]*(cMat[i][j] - 1)*(3*c-1)*(c-2)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/24.0;
		}
	}
	
	
	term[5] = c*(6*a*b + (c-1)*(3*a + 3*b + 2*c - 1))*ESF(a_cA,tA,P)*ESF(b_cB,tB,Q)/3.0;
	
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		
		if (!P.empty()){
			term[6] = term[6] - (c-1)*tA*cA[i]*P[i]*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q)/2.0;
		}
		else if (aVec[i] == 0 && cA[i] == 1) {
			term[6] = term[6] - (c-1)*tA*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q)/2.0;
		}
	}
	
	
	for(j=0; j< L; j++){
		tempmatrix = b_cB;
		tempmatrix[j] = tempmatrix[j] - 1;
		
		if (!Q.empty()){
			term[7] = term[7] - (c-1)*tB*cB[j]*Q[j]*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q)/2.0;
		}
		else if (bVec[j] == 0 && cB[j] == 1) {
			term[7] = term[7] - (c-1)*tB*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q)/2.0;
		}
		
	}
	
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		term[8] = term[8] - cA[i]*(cA[i] - 1)*((c+1)*(3*c-2) - 6*(tA + 2*(a+c-1)))*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q)/24.0;
	}
	
	
	for(i=0; i< K; i++){
		for(k=0; k< K; k++){
			tempmatrix = a_cA;
			tempmatrix[i] = tempmatrix[i] - 1;
			tempmatrix[k] = tempmatrix[k] - 1;
			if (i == k){
				term[9] = term[9] - cA[i]*(cA[i] - 1)*(aVec[i] + cA[i] - 1)*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q)/2.0;
				term[9] = term[9] + ((cA[i]*(cA[i] - 1)*(cA[i] - 2)*(cA[i] - 3)/8.0) + (4*cA[i]*(cA[i] - 1)*(cA[i]-2)/6.0) + (cA[i]*(cA[i] - 1)/2.0))*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q);
			}else{
				
				term[9] = term[9] + cA[i]*(cA[i] - 1)*cA[k]*(cA[k] - 1)*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q)/8.0;
			}
		}
	}
	
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		for(j=0; j< L; j++){
			tempmatrix2 = b_cB;
			tempmatrix2[j] = tempmatrix2[j] - 1;
			term[10] = term[10] + ((cA[i]-1)*cA[i]*(cB[j]-1)*cB[j]/8.0 + cMat[i][j]*(1+2*cA[i]*cB[j]-2*cA[i]-2*cB[j])/12.0 + (cMat[i][j]*cMat[i][j])/12.0)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q);
		}
	}
	
	for(i=0; i< K; i++){
		for(j=0; j< L; j++){
			tempmatrix2 = b_cB;
			tempmatrix2[j] = tempmatrix2[j] - 1;
			for(k=0; k< K; k++){
				tempmatrix = a_cA;
				tempmatrix[i] = tempmatrix[i] - 1;
				tempmatrix[k] = tempmatrix[k] - 1;
				if (i == k){
					term[11] = term[11] - cMat[i][j]*(cMat[i][j]-1)*(1-2*cA[i])*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/6.0;
					term[11] = term[11] - cMat[i][j]*(cMat[i][j]-1)*(cA[i]+1)*(3*cA[i]-2)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/24.0;
				}else{
					term[11] = term[11] - cMat[i][j]*(cMat[i][j]-1)*cA[k]*(cA[k]-1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/8.0;
				}
			}
		}
	}
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		term[12] = term[12] - 2*cA[i]*((2*b+c-1)*(aVec[i]+cA[i]-1)/2.0 - (cA[i]-1)*(c-2+3*b)/6.0)*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q);
	}
    
	for(i=0; i< K; i++){
			tempmatrix = a_cA;
			tempmatrix[i] = tempmatrix[i] - 2;
				
		if (!P.empty())
			term[13] = term[13] + tA*P[i]*cA[i]*(cA[i] - 1)*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q)/4.0;
		else if (cA[i] == 2)
			term[13] = term[13] + tA*cA[i]*(cA[i] - 1)*ESF(tempmatrix,tA,P)*ESF(b_cB,tB,Q)/4.0;
	}
	
	for(j=0; j< L; j++){
		tempmatrix = b_cB;
		tempmatrix[j] = tempmatrix[j] - 1;
		term[14] = term[14] - cB[j]*(cB[j] - 1)*((c+1)*(3*c-2) - 6*(tB + 2*(b+c-1)))*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q)/24.0;
	}
	
	term[15] = term[10];
	
	for(j=0; j< L; j++){
		for(l=0; l< L; l++){
			tempmatrix = b_cB;
			tempmatrix[j] = tempmatrix[j] - 1;
			tempmatrix[l] = tempmatrix[l] - 1;
			if (j == l){
				term[16] = term[16] - cB[j]*(cB[j] - 1)*(bVec[j] + cB[j] - 1)*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q)/2.0;
				term[16] = term[16] + ((cB[j]*(cB[j] - 1)*(cB[j] - 2)*(cB[j] - 3)/8.0) + (4*cB[j]*(cB[j] - 1)*(cB[j]-2)/6.0) + (cB[j]*(cB[j] - 1)/2.0))*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q);
			}else{
				term[16] = term[16] + cB[j]*(cB[j] - 1)*cB[l]*(cB[l] - 1)*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q)/8.0;
			}
		}
	}
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		for(j=0; j< L; j++){
			for(l=0; l< L; l++){
				tempmatrix2 = b_cB;
				tempmatrix2[j] = tempmatrix2[j] - 1;
				tempmatrix2[l] = tempmatrix2[l] - 1;
				if (j == l){
					term[17] = term[17] - cMat[i][j]*(cMat[i][j]-1)*(1-2*cB[j])*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/6.0;
					term[17] = term[17] - cMat[i][j]*(cMat[i][j]-1)*(cB[j]+1)*(3*cB[j]-2)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/24.0;
				}else{
					term[17] = term[17] - cMat[i][j]*(cMat[i][j]-1)*cB[l]*(cB[l]-1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/8.0;
				}
			}
		}
	}
	
	for(j=0; j< L; j++){
		tempmatrix = b_cB;
		tempmatrix[j] = tempmatrix[j] - 1;
		term[18] = term[18] - 2*cB[j]*((2*a+c-1)*(bVec[j]+cB[j]-1)/2.0 - (cB[j]-1)*(c-2+3*a)/6.0)*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q);
	}
	
	for(j=0; j< L; j++){
		tempmatrix = b_cB;
		tempmatrix[j] = tempmatrix[j] - 2;
		
		if (!Q.empty())
			term[19] = term[19] + tB*Q[j]*cB[j]*(cB[j] - 1)*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q)/4.0;
		else if (cB[j] == 2)
			term[19] = term[19] + tB*cB[j]*(cB[j] - 1)*ESF(a_cA,tA,P)*ESF(tempmatrix,tB,Q)/4.0;
		
	}
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		for(j=0; j< L; j++){
			tempmatrix2 = b_cB;
			tempmatrix2[j] = tempmatrix2[j] - 1;
			term[20] = term[20] + cMat[i][j]*(cMat[i][j]-1)*((c+1)*(3*c-2)/24.0 - (tA + tB + 2*(a+c-1) + 2*(b+c-1))/4.0)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q);
		}
	}
	
	for(i=0; i< K; i++){
		for(j=0; j< L; j++){
			tempmatrix2 = b_cB;
			tempmatrix2[j] = tempmatrix2[j] - 1;
			for(k=0; k< K; k++){
				tempmatrix = a_cA;
				tempmatrix[i] = tempmatrix[i] - 1;
				tempmatrix[k] = tempmatrix[k] - 1;
				if (i == k){
					term[21] = term[21] + cMat[i][j]*(cMat[i][j]-1)*(aVec[i] + cA[i] - 1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/2.0;
					term[21] = term[21] - cMat[i][j]*(cMat[i][j]-1)*(cA[i]+1)*(3*cA[i]-2)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/24.0;
				}
				else{
					term[21] = term[21] - cMat[i][j]*(cMat[i][j]-1)*cA[k]*(cA[k]-1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/8.0;
				}
			}
		}
	}
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		for(j=0; j< L; j++){
			for(l=0; l< L; l++){
				tempmatrix2 = b_cB;
				tempmatrix2[j] = tempmatrix2[j] - 1;
				tempmatrix2[l] = tempmatrix2[l] - 1;
				if (j == l){
					term[22] = term[22] + cMat[i][j]*(cMat[i][j]-1)*(bVec[j] + cB[j] - 1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/2.0;
					term[22] = term[22] - cMat[i][j]*(cMat[i][j]-1)*(cB[j]+1)*(3*cB[j]-2)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/24.0;
				}
				else{
					term[22] = term[22] - cMat[i][j]*(cMat[i][j]-1)*cB[l]*(cB[l]-1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/8.0;
				}
			}
		}
	}
	
	
	for(i=0; i< K; i++){
		for(j=0; j< L; j++){
			for(k=0; k< K; k++){
				tempmatrix = a_cA;
				tempmatrix[i] = tempmatrix[i] - 1;
				tempmatrix[k] = tempmatrix[k] - 1;
				for(l=0; l< L; l++){
					tempmatrix2 = b_cB;
					tempmatrix2[j] = tempmatrix2[j] - 1;
					tempmatrix2[l] = tempmatrix2[l] - 1;
					if (i == k && j == l)
						term[23] = term[23] + cMat[i][j]*(cMat[i][j]-1)*(cMat[i][j]-2)*(3*cMat[i][j]-1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/24.0;
					else
						term[23] = term[23] + cMat[i][j]*(cMat[i][j]-1)*cMat[k][l]*(cMat[k][l]-1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/8.0;
				}
			}
		}
	}
	
	for(i=0; i< K; i++){
			tempmatrix = a_cA;
			tempmatrix[i] = tempmatrix[i] - 2;
			for(j=0; j< L; j++){
				tempmatrix2 = b_cB;
				tempmatrix2[j] = tempmatrix2[j] - 1;
							
			if(!P.empty())
				term[24] = term[24] - tA*P[i]*cMat[i][j]*(cMat[i][j] - 1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/4.0;
			else if (aVec[i] + cA[i] == 2)
				term[24] = term[24] - tA*cMat[i][j]*(cMat[i][j] - 1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/4.0;
			
			}
		}	
	
	for(j=0; j< L; j++){
			tempmatrix2 = b_cB;
			tempmatrix2[j] = tempmatrix2[j] - 2;
			for(i=0; i< K; i++){
				tempmatrix = a_cA;
				tempmatrix[i] = tempmatrix[i] - 1;
						
			if(!Q.empty())
				term[25] = term[25] - tB*Q[j]*cMat[i][j]*(cMat[i][j] - 1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/4.0;
			else if (bVec[j] + cB[j] == 2)
				term[25] = term[25] - tB*cMat[i][j]*(cMat[i][j] - 1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/4.0;
			
			}
		}
	
	for(i=0; i< K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		for(j=0; j< L; j++){
			tempmatrix2 = b_cB;
			tempmatrix2[j] = tempmatrix2[j] - 1;
			term[26] = term[26] + 2*((aVec[i] + cA[i])*(bVec[j] + cB[j])*cMat[i][j] - (aVec[i] + cA[i])*cMat[i][j]*(cB[j]+1)/2.0 - (bVec[j] + cB[j])*cMat[i][j]*(cA[i]+1)/2.0 + cMat[i][j]*(2*cA[i]*cB[j] + cA[i] + cB[j] + cMat[i][j] + 1)/6.0)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q);
		}
	}
	
	for (i=0; i<K; i++){
		tempmatrix = a_cA;
		tempmatrix[i] = tempmatrix[i] - 1;
		for (j=0; j<L; j++){
			tempmatrix2 = b_cB;
			tempmatrix2[j] = tempmatrix2[j] - 1;
						
			if (!P.empty())
				term[27] = term[27] + tB*Q[j]*cMat[i][j]*(cA[i]-1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/2.0;
			else if (cMat[i][j] == 1 && bVec[j] == 0 && cB[j] == 1)
				term[27] = term[27] + tB*(cA[i] - 1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/2.0;
			
			if (!Q.empty())
				term[28] = term[28] + tA*P[i]*cMat[i][j]*(cB[j]-1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/2.0;
			else if (cMat[i][j] == 1 && aVec[i] == 0 && cA[i] == 1)
				term[28] = term[28] + tA*(cB[j] - 1)*ESF(tempmatrix,tA,P)*ESF(tempmatrix2,tB,Q)/2.0;
			
			}
	}
		
	
  return total(term);
}
