#include <math.h>
#include <map>
#include <stdio.h>
#include <ctype.h>
#include <algorithm>
#include <utility>
#include <vector>
#include "getq.h"
#include "basicdefs.h"
#include "weakComposition.h"
#include "getSubs.h"

using namespace std;

// This function computes q_2(a+cA,b+cB,0) 
// This is an implementation using weak integer compositions, which can be exploited in the infinite-alleles case, or the PIM case provided P_i = Q_j for all i,j.
double qtwoRec(vector<int>& aVec, vector<int>& bVec, vector<vector<int> >& cMat,double thetaA, double thetaB, vector<double>& P, vector<double>& Q){
  
	bool use_weak = P.empty(); // An indicator for whether we can use weak integer compositions
	if (!P.empty())
	{
		use_weak = (P.size() == Q.size());
		for (int i=0; i<P.size(); i++)
		{
			for (int j=0; j<Q.size(); j++)
			{
				use_weak &= (P[i] == Q[j]);
			}
		}
	}
	
  vector<int> cA = marginalVec(cMat,'A');
  vector<int> cB = marginalVec(cMat,'B');
  vector<int> a_cA = aVec+cA;
  vector<int> b_cB = bVec+cB;
  int Total = total(a_cA)+total(b_cB);
  int ATotal = total(cA)+total(aVec);
  int BTotal = total(cB)+total(bVec);
  double a,b,normalization;
  
  typedef map<vector<int>, double> mapType;
  vector<mapType> DPT(2); // DP Table
  int iT,iI,iJ,iCA,iCB,iA,iB;
  bool parity=0, newparity=1;
  vector<vector<int> > currentA, currentB;
  double tmp;
  int ai1Count, bi1Count;
	double ai1CountPIM, bi1CountPIM;
	
  //initialize DP table;
  vector<int> tempmatrix(aVec.size()+bVec.size());
  fill(tempmatrix.begin(),tempmatrix.end(),0);
  for(iI=0; iI < tempmatrix.size(); iI++){
    tempmatrix[iI]=1;
    DPT[0].insert(pair<vector<int>, double>(tempmatrix,0.0));
    tempmatrix[iI]=0;
  }

  int limit;
  for(iT=2; iT <= Total; iT++)
  {
	  // iA + iB = iT, where iA > 0 and iB > 0
	  limit = (iT > ATotal)?  ATotal+1: iT;
	  for(iA = 2; iA < limit; iA++)
	  {
		  iB=iT-iA;
		  if (use_weak) currentA = getWeakComp_n_k(a_cA,iA);
		  else currentA = getSubsamples(a_cA,iA);
		
		  if(iB <= BTotal)
		  {
			  if (use_weak) currentB = getWeakComp_n_k(b_cB,iB);
			  else currentB = getSubsamples(b_cB,iB);
			  
			  for(iCA=0; iCA < currentA.size(); iCA++)
			  {
				  for(iCB=0; iCB < currentB.size(); iCB++)
				  {
					  if(iB <= 1 || iA <=1)
					  {
						  DPT[newparity].insert(pair<vector<int>, double>(concatenate(currentA[iCA],currentB[iCB]),0.0));
					  }
					  else if (P.empty())
					  {
						  tmp=0.0; ai1Count=0; bi1Count=0;
						  for(iI=0; iI < aVec.size(); iI++)
						  {
							  if(currentA[iCA][iI] > 1){
								  currentA[iCA][iI] -= 1;
								  tmp += currentA[iCA][iI]*(currentA[iCA][iI]+1)*DPT[parity][concatenate(currentA[iCA],currentB[iCB])];
								  currentA[iCA][iI] += 1;
							  }
							  else{
								  if(currentA[iCA][iI] == 1){
									  ai1Count++;
									  currentA[iCA][iI] -= 1;
									  tmp += thetaA*DPT[parity][concatenate(currentA[iCA],currentB[iCB])];
									  currentA[iCA][iI] += 1;
								  }
							  }
						  }
						  for(iJ=0; iJ < bVec.size(); iJ++)
						  {
							  if(currentB[iCB][iJ] > 1){
								  currentB[iCB][iJ] -= 1;
								  tmp += currentB[iCB][iJ]*(currentB[iCB][iJ]+1)*DPT[parity][concatenate(currentA[iCA],currentB[iCB])];
								  currentB[iCB][iJ] += 1;
							  }
							  else{
								  if(currentB[iCB][iJ] == 1){
									  bi1Count++;
									  currentB[iCB][iJ] -= 1;
									  tmp += thetaB*DPT[parity][concatenate(currentA[iCA],currentB[iCB])];
									  currentB[iCB][iJ] += 1;
								  }
							  }
						  }
						  
						  a=total(currentA[iCA]);
						  b=total(currentB[iCB]);
						  if(a > 1 && b > 1) tmp += 4*ESF(currentA[iCA],thetaA,P)*ESF(currentB[iCB],thetaB,Q)*(a*thetaA-(thetaA+a-1)*ai1Count)*(b*thetaB-(thetaB+b-1)*bi1Count);
						  normalization=a*(a-1+thetaA)+b*(b-1+thetaB);
						  DPT[newparity].insert(pair<vector<int>, double>(concatenate(currentA[iCA],currentB[iCB]),tmp/normalization));
					  }
					  else
					  {
						  tmp=0.0; ai1CountPIM=0.0; bi1CountPIM=0.0;
						  for(iI=0; iI < aVec.size(); iI++){
							  if(currentA[iCA][iI] > 0){
								  currentA[iCA][iI] -= 1;
								  tmp += (currentA[iCA][iI]+1)*(currentA[iCA][iI] + thetaA*P[iI])*DPT[parity][concatenate(currentA[iCA],currentB[iCB])]; // (All entries in P should be the same)
								  currentA[iCA][iI] += 1;
								  ai1CountPIM += currentA[iCA][iI]*(currentA[iCA][iI]-1)/(thetaA*P[iI] + currentA[iCA][iI] - 1);
							  }
						  }
						  for(iJ=0; iJ < bVec.size(); iJ++){
							  if(currentB[iCB][iJ] > 0){
								  currentB[iCB][iJ] -= 1;
								  tmp += (currentB[iCB][iJ]+1)*(currentB[iCB][iJ] + thetaB*Q[iJ])*DPT[parity][concatenate(currentA[iCA],currentB[iCB])];
								  currentB[iCB][iJ] += 1;
								  bi1CountPIM += currentB[iCB][iJ]*(currentB[iCB][iJ]-1)/(thetaB*Q[iJ] + currentB[iCB][iJ] - 1);
							  }
						  }

						  a=total(currentA[iCA]);
						  b=total(currentB[iCB]);
						  if(a > 1 && b > 1)	tmp += 4*ESF(currentA[iCA],thetaA,P)*ESF(currentB[iCB],thetaB,Q)*(a*(a-1)-(thetaA+a-1)*ai1CountPIM)*(b*(b-1)-(thetaB+b-1)*bi1CountPIM);

						  normalization=a*(a-1+thetaA)+b*(b-1+thetaB);
						  DPT[newparity].insert(pair<vector<int>, double>(concatenate(currentA[iCA],currentB[iCB]),tmp/normalization));
					  }
				  }
			  }
		  }
	  }
	
    DPT[parity].clear();
    parity = (parity+1)%2;
    newparity = (newparity+1)%2;
  }
  return DPT[parity][concatenate(a_cA,b_cB)];
}


// This function computes q_2(a+cA,b+cB,0) 
// This is an implementaion using a straightforward DP table
double qtwoRec2(vector<int>& aVec, vector<int>& bVec, vector<vector<int> >& cMat,double thetaA, double thetaB, vector<double>& P, vector<double>& Q){
  vector<int> cA = marginalVec(cMat,'A');
  vector<int> cB = marginalVec(cMat,'B');
  vector<int> a_cA = aVec+cA;
  vector<int> b_cB = bVec+cB;
  double a,b,normalization,tmp;
  
  vector<vector<vector<vector<double> > > > DPT; // DP Table
  int iI,iJ,iK,iL;
  int ai1Count, bi1Count;
	double ai1CountPIM, bi1CountPIM;
  vector<int> tmpAVec(2), tmpBVec(2);

  if(a_cA.size() == 1) a_cA.push_back(0);
  if(b_cB.size() == 1) b_cB.push_back(0);

  DPT.resize(a_cA[0]+1);
  for(iI=0; iI <= a_cA[0]; iI++){
    DPT[iI].resize(a_cA[1]+1);
    for(iJ=0; iJ <= a_cA[1]; iJ++){
      DPT[iI][iJ].resize(b_cB[0]+1);
      for(iK=0; iK <= b_cB[0]; iK++){
		  DPT[iI][iJ][iK].resize(b_cB[1]+1);
		  for(iL=0; iL <= b_cB[1]; iL++){
			  a=iI+iJ;
			  b=iK+iL;
			  if(a <= 1 || b <= 1){
				  DPT[iI][iJ][iK][iL]=0.0;
			  }
			  else if (P.empty()){
				  tmp=0.0;
				  ai1Count=0;
				  bi1Count=0;
				  if(iI > 1){
					  tmp += iI*(iI-1)*DPT[iI-1][iJ][iK][iL];
				  }
				  else{
					  if(iI==1){
						  ai1Count++;
						  tmp += thetaA*DPT[iI-1][iJ][iK][iL];
					  }
				  }
				  if(iJ > 1){
					  tmp += iJ*(iJ-1)*DPT[iI][iJ-1][iK][iL];
				  }
				  else{
					  if(iJ==1){
						  ai1Count++;
						  tmp += thetaA*DPT[iI][iJ-1][iK][iL];
					  }
				  }
				  if(iK > 1){
					  tmp +=iK*(iK-1)*DPT[iI][iJ][iK-1][iL];
				  }
				  else{
					  if(iK==1){
						  bi1Count++;
						  tmp += thetaB*DPT[iI][iJ][iK-1][iL];
					  }
				  }
				  if(iL > 1){
					  tmp +=iL*(iL-1)*DPT[iI][iJ][iK][iL-1];
				  }
				  else{
					  if(iL==1){
						  bi1Count++;
						  tmp += thetaB*DPT[iI][iJ][iK][iL-1];
					  }
				  }
				  tmpAVec[0]=iI; tmpAVec[1]=iJ;
				  tmpBVec[0]=iK; tmpBVec[1]=iL;
				  if(a > 1 && b > 1){
					  tmp += 4*ESF(tmpAVec,thetaA,P)*ESF(tmpBVec,thetaB,Q)*(a*thetaA-(thetaA+a-1)*ai1Count)*(b*thetaB-(thetaB+b-1)*bi1Count);
				  }
				  normalization=a*(a-1+thetaA)+b*(b-1+thetaB);
				  DPT[iI][iJ][iK][iL]=tmp/normalization;
			  }
			  else{
				  tmp=0.0;
				  ai1CountPIM=0.0;
				  bi1CountPIM=0.0;
				  if (iI > 0){
					  tmp += iI*(iI - 1 + thetaA*P[0])*DPT[iI-1][iJ][iK][iL];
					  ai1CountPIM += iI*(iI-1)/(thetaA*P[0] + iI - 1);
				  }
				  if (iJ > 0){
					  tmp += iJ*(iJ - 1 + thetaA*P[1])*DPT[iI][iJ-1][iK][iL];
					  ai1CountPIM += iJ*(iJ-1)/(thetaA*P[1] + iJ - 1);
				  }
				  if (iK > 0){
					  tmp += iK*(iK - 1 + thetaB*Q[0])*DPT[iI][iJ][iK-1][iL];
					  bi1CountPIM += iK*(iK-1)/(thetaB*Q[0] + iK - 1);
				  }
				  if (iL > 0){
					  tmp += iL*(iL - 1 + thetaB*Q[1])*DPT[iI][iJ][iK][iL-1];
					  bi1CountPIM += iL*(iL-1)/(thetaB*Q[1] + iL - 1);
				  }
				  tmpAVec[0]=iI; tmpAVec[1]=iJ;
				  tmpBVec[0]=iK; tmpBVec[1]=iL;
				  if(a > 1 && b > 1){
					  tmp += 4*ESF(tmpAVec,thetaA,P)*ESF(tmpBVec,thetaB,Q)*(a*(a-1)-(thetaA+a-1)*ai1CountPIM)*(b*(b-1)-(thetaB+b-1)*bi1CountPIM);
				  }
				  normalization=a*(a-1+thetaA)+b*(b-1+thetaB);
				  DPT[iI][iJ][iK][iL]=tmp/normalization;
			  }
			  
		  }
      }
    }
  }

  return DPT[a_cA[0]][a_cA[1]][b_cB[0]][b_cB[1]];
}
